#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A wrapper around multi-stage ThreatExchange operations.

Includes simple matching and writing back. Useful for quickly validating new
sources of ThreatExchange data. A possible template for a native
implementation in your own architecture.

This helper heavily relies on a config file to provide consistent behavior
between stages, and a state file to store hashes.
"""

import argparse
import inspect
import os
import os.path
import pathlib
import sys
import typing as t

from .. import TE
from . import command_base as base, fetch, experimental_fetch, label, match
from ..collab_config import CollaborationConfig
from ..dataset import Dataset


def get_subcommands() -> t.List[t.Type[base.Command]]:
    return [
        fetch.FetchCommand,
        experimental_fetch.ExperimentalFetchCommand,
        match.MatchCommand,
        label.LabelCommand,
    ]


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    ap.add_argument(
        "--config",
        "-c",
        type=argparse.FileType("r"),
        help="a ThreatExchange collaboration config",
    )
    ap.add_argument(
        "--app-token",
        "-a",
        type=CollaborationConfig.load,
        metavar="TOKEN",
        help="the App token for ThreatExchange",
    )
    ap.add_argument(
        "--state-dir",
        "-s",
        type=_verify_directory,
        metavar="DIR",
        help="the directory with the config state",
    )
    subparsers = ap.add_subparsers(title="verbs", help="which action to do")
    for command in get_subcommands():
        command.add_command_to_subparser(subparsers)

    return ap


def execute_command(namespace) -> None:
    if not hasattr(namespace, "command_cls"):
        get_argparse().print_help()
        return
    command_cls = namespace.command_cls
    try:
        # Init TE lib
        init_app_token(namespace.app_token)
        # Init collab config
        cfg = init_config_file(namespace.config)
        # "Init" dataset
        dataset = Dataset(cfg, namespace.state_dir)
        command_argspec = inspect.getfullargspec(command_cls.__init__)
        arg_names = set(command_argspec[0])
        # Since we didn't import click, use hard-to-debug magic to init the command
        command = command_cls(
            **{k: v for k, v in namespace.__dict__.items() if k in arg_names}
        )
        command.execute(dataset)
    except base.CommandError as ce:
        print(ce, file=sys.stderr)
        sys.exit(ce.returncode)
    except KeyboardInterrupt:
        # No stack for CTRL+C
        sys.exit(130)


def init_app_token(cli_option: str = None) -> None:
    """Initialize the API key from a variety of fallback sources"""

    file_loc = pathlib.Path("~/.txtoken").expanduser()
    environment_var = "TX_ACCESS_TOKEN"
    if cli_option:
        TE.Net.APP_TOKEN = cli_option
    elif os.environ.get(environment_var):
        TE.Net.APP_TOKEN = os.environ[environment_var]
    elif file_loc.exists() and file_loc.read_text():
        TE.Net.APP_TOKEN = file_loc.read_text()
    else:
        raise base.CommandError(
            (
                "Can't find API key - pass as an argument, in the environment as "
                f"{environment_var} or put it in {file_loc}"
            ),
            2,
        )


def init_config_file(cli_provided: t.IO = None) -> CollaborationConfig:
    """Initialize the collaboration file from a variety of sources"""
    if cli_provided is not None:
        return CollaborationConfig.load(cli_provided)
    path_order = ("te.cfg", "~/te.cfg")
    for loc in path_order:
        path = pathlib.Path(loc).expanduser()
        if path.exists():
            break
    else:
        print(
            (
                "Looks like you haven't set up a collaboration config, "
                "so using the sample one against public data"
            ),
            file=sys.stderr,
        )
        return CollaborationConfig.get_example_config()
    with path.open() as f:
        return CollaborationConfig.load(f)


def _verify_directory(raw: str) -> pathlib.Path:
    ret = pathlib.Path(raw)
    if ret.exists() and not ret.is_dir():
        raise argparse.ArgumentTypeError(f"{ret} is a file, not a directory")
    return ret


def main() -> None:
    ap = get_argparse()
    namespace = ap.parse_args()
    execute_command(namespace)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrappers for the json returned by the ThreatExchange API to typed objects.
"""

import typing as t

from . import TE


class ThreatIndicator(t.NamedTuple):
    """
    Wrapper around ThreatExchange JSON for a ThreatIndicator.

    Example:
    {
      "id": "3058061737574159",
      "indicator": "facefacefacefacefacefacefaceface",
      "type": "HASH_PDQ",
      "creation_time": 1601513795,
      "last_updated": 1601513796,
      "status": "MALICIOUS",
      "is_expired": false,
      "tags": [
        "tag1",
        "tag2",
        "tag3"
      ],
      "applications_with_opinions": [
        "123456789055502"
      ],
      "expire_time": 9601513796
    }
    """

    id: int
    indicator: str
    threat_type: str
    creation_time: int
    last_updated: int
    status: str
    is_expired: bool
    tags: t.List[str]
    applications_with_opinions: t.List[int]
    expire_time: int

    def as_row(self) -> t.Tuple[int, str, str, int, int, str, bool, str, str, int]:
        """Simple conversion to CSV row"""
        return (
            self.id,
            self.indicator,
            self.threat_type,
            self.creation_time,
            self.last_updated,
            self.status,
            self.is_expired,
            " ".join(self.tags),
            " ".join([str(app) for app in self.applications_with_opinions]),
            self.expire_time,
        )

    @classmethod
    def from_row(self, row: t.Iterable) -> "ThreatIndicator":
        """Simple conversion from CSV row"""
        tags = row[7].split(" ") if row[7] else []
        apps = [int(app) for app in (row[8].split(" ") if row[8] else [])]
        expire_time = int(row[9]) if row[9] else None
        return ThreatIndicator(
            int(row[0]),
            row[1],
            row[2],
            int(row[3]),
            int(row[4]),
            row[5],
            row[6] == "True",
            tags,
            apps,
            expire_time,
        )



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import collections
import pathlib
import time
import typing as t
from .. import TE
from ..dataset import Dataset
from ..indicator import ThreatIndicator
from ..content_type import meta
from . import command_base


class ExperimentalFetchCommand(command_base.Command):
    """
    WARNING: This is experimental, you probably want to use "Fetch" instead.

    Download content from ThreatExchange to disk.

    Using the CollaborationConfig, identify ThreatPrivacyGroup that
    corresponds to a single collaboration and fetch related threat updates.
    """

    PROGRESS_PRINT_INTERVAL_SEC = 30

    @classmethod
    def init_argparse(cls, ap) -> None:
        ap.add_argument(
            "--start-time",
            type=int,
            help="Fetch updates that occured on or after this timestamp",
        )
        ap.add_argument(
            "--stop-time",
            type=int,
            help="Fetch updates that occured before this timestamp",
        )
        ap.add_argument(
            "--owner",
            type=int,
            help="Only fetch updates for indicators that the given app has a descriptor for",
        )
        ap.add_argument(
            "--threat-types",
            nargs="+",
            help="Only fetch updates for indicators of the given type",
        )
        ap.add_argument(
            "--additional-tags",
            nargs="+",
            help="Only fetch updates for indicators that have a descriptor with each of these tags",
        )

    def __init__(
        self,
        start_time: int,
        stop_time: int,
        owner: int,
        threat_types: t.List[str],
        additional_tags: t.List[str],
    ) -> None:
        self.start_time = start_time
        self.stop_time = stop_time
        self.owner = owner
        self.threat_types = threat_types
        self.additional_tags = additional_tags
        self.signal_types_by_name = {
            name: signal() for name, signal in meta.get_signal_types_by_name().items()
        }
        self.last_update_printed = 0
        self.counts = collections.Counter()

    def execute(self, dataset: Dataset) -> None:
        # TODO: [Potential] Force full fetch if it has been 90 days since the last fetch.
        self.start_time = (
            dataset.get_indicator_checkpoint()
            if self.start_time is None
            else self.start_time
        )
        self.stop_time = int(time.time()) if self.stop_time is None else self.stop_time
        dataset.load_indicator_cache(self.signal_types_by_name.values())
        more_to_fetch = True
        next_page = None

        while more_to_fetch:
            result = TE.Net.getThreatUpdates(
                dataset.config.privacy_groups[0],
                start_time=self.start_time,
                stop_time=self.stop_time,
                owner=self.owner,
                threat_type=self.threat_types,
                additional_tags=self.additional_tags,
                next_page=next_page,
            )
            if "data" in result:
                self._process_indicators(result["data"])
            more_to_fetch = "paging" in result and "next" in result["paging"]
            next_page = result["paging"]["next"] if more_to_fetch else None

        for signal_name, signal_type in self.signal_types_by_name.items():
            if signal_name not in self.counts:
                continue
            dataset.store_indicator_cache(signal_type)
            print(f"{signal_name}: {self.counts[signal_name]}")

        dataset.record_indicator_checkpoint(self.stop_time)

    def _process_indicators(
        self,
        indicators: list,
    ) -> None:
        """Process indicators"""
        for ti_json in indicators:
            ti = ThreatIndicator(
                int(ti_json.get("id")),
                ti_json.get("indicator"),
                ti_json.get("type"),
                int(ti_json.get("creation_time")),
                int(ti_json.get("last_updated")),
                ti_json.get("status"),
                ti_json.get("is_expired"),
                ti_json.get("tags"),
                [int(app) for app in ti_json.get("applications_with_opinions")],
                int(ti_json.get("expire_time")) if "expire_time" in ti_json else None,
            )

            match = False
            for signal_name, signal_type in self.signal_types_by_name.items():
                if signal_type.process_indicator(ti):
                    match = True
                    self.counts[signal_name] += 1
            if match:
                self.counts["all"] += 1

        now = time.time()
        if now - self.last_update_printed >= self.PROGRESS_PRINT_INTERVAL_SEC:
            self.last_update_printed = now
            self.stderr(f"Processed {self.counts['all']}...")



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrappers for the json returned by the ThreatExchange API to typed objects.
"""

import collections
import typing as t

from . import TE


class ThreatDescriptor(t.NamedTuple):
    """
    Wrapper around ThreatExchange JSON for a ThreatDescriptor.

    Example:
    {
        "id": "3058061737574159"
        "raw_indicator": "facefacefacefacefacefacefaceface",
        "type": "HASH_MD5",
        "owner": {
            "id": "616912251743987",
            ...,
        },
        "status": "MALICIOUS",
        "tags": null,
        "added_on": "2020-07-01T18:31:15+0000"
        ...
    }
    """

    # You declared the indicator was in the collaboration label set
    TRUE_POSITIVE = "true_positive"
    # You declared the indicator was not in the collaboration label set
    FALSE_POSITIVE = "false_positive"
    # Someone declared the indicator was not in the collaboration label set
    DISPUTED = "disputed"

    # Special tags to mark whether you (or someone else)
    # has weighed in on the indicator
    SPECIAL_TAGS = frozenset((TRUE_POSITIVE, FALSE_POSITIVE, DISPUTED))

    id: int
    raw_indicator: str
    indicator_type: str
    owner_id: int
    tags: t.List[str]
    status: str
    added_on: str

    def to_params(self) -> t.Dict[str, t.Any]:
        params = dict(self.__dict__)
        params["type"] = params.pop("indicator_type")
        if not params["tags"]:
            del params["tags"]
        return params

    @property
    def is_true_positive(self) -> bool:
        return self.TRUE_POSITIVE in self.tags

    @property
    def is_false_positive(self) -> bool:
        return self.FALSE_POSITIVE in self.tags

    @property
    def is_mine(self) -> bool:
        """This Descriptor is my App's Opinion"""
        # TODO - come up with a way to do this that doesn't use class state
        return TE.Net.APP_TOKEN.partition("|")[0] == str(self.owner_id)


class SimpleDescriptorRollup:
    """
    A simple way to merge opinions on the same indicator.

    This contains all the information needed for simple SignalType state.
    """

    IS_MY_OPINION = "mine"

    __slots__ = ["first_descriptor_id", "added_on", "labels"]

    def __init__(
        self, first_descriptor_id: int, added_on: str, labels: t.Set[str]
    ) -> None:
        self.first_descriptor_id = first_descriptor_id
        self.added_on = added_on  # TODO - convert to int?
        self.labels = set(labels)

    @classmethod
    def from_descriptor(cls, descriptor: ThreatDescriptor) -> "SimpleDescriptorRollup":
        return cls(descriptor.id, descriptor.added_on, descriptor.tags)

    def merge(self, descriptor: ThreatDescriptor) -> None:
        # Is the other descriptor mine? If so, unconditionally take it and clear
        # everything else
        if descriptor.is_mine:
            self.added_on = self.IS_MY_OPINION
            self.first_descriptor_id = descriptor.id
            self.labels = descriptor.tags
            return
        # My descriptor beats my reactions, and I don't want
        # to take anyone else's opinion
        elif self.added_on == self.IS_MY_OPINION:
            return
        # Is my reaction?
        elif descriptor.is_false_positive:
            self.added_on = self.IS_MY_OPINION
            self.first_descriptor_id = descriptor.id
            self.labels = descriptor.tags
            return
        # Else merge the labels together
        self.added_on, self.first_descriptor_id = min(
            (self.added_on, self.first_descriptor_id),
            (descriptor.added_on, descriptor.id),
        )
        self.labels.union(descriptor.tags)

    def as_row(self) -> t.Tuple[int, str, str]:
        """Simple conversion to CSV row"""
        return self.first_descriptor_id, self.added_on, " ".join(self.labels)

    @classmethod
    def from_row(cls, row: t.Iterable) -> "SimpleDescriptorRollup":
        """Simple conversion from CSV row"""
        labels = []
        if row[2]:
            labels = row[2].split(" ")
        return cls(int(row[0]), row[1], labels)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Core abstractions for signal types.
"""

import csv
import pathlib
import re
import typing as t

from .. import common
from ..descriptor import SimpleDescriptorRollup, ThreatDescriptor
from ..indicator import ThreatIndicator


class SignalMatch(t.NamedTuple):
    # TODO - Labels probably don't belong here, because we just duplicate storage
    #        better to have descriptor lookup with the full labels
    labels: t.Set[str]
    primary_descriptor_id: int
    # TODO: Someday, also include the following:
    # contested: bool
    # plurality_opinion: t.Tuple(bool, t.Set[str])
    # highest_action_severity: str
    # opinions_by_owner: t.Dict[int, MatchOpinion]


class SignalType:
    """
    Abstraction for different signal types.

    A signal type is an intermediate representation of content that can be used
    to match against similar or identical content. Sometimes called a "hash"
    type.

    This class additionally helps translates ThreatDescriptors into the correct
    representation to do matching, as well as serialize that representation
    into a compact form.
    """

    @classmethod
    def get_name(cls):
        """A compact name in lower_with_underscore style (used in filenames)"""
        return common.class_name_to_human_name(cls.__name__, "Signal")

    def process_descriptor(self, descriptor: t.Dict[str, t.Any]) -> bool:
        """
        Add ThreatDescriptor to the state of this type, if it is for this type.

        Return true if the true if the descriptor was used.
        """
        return False

    def process_indicator(self, indicator: t.Dict[int, t.Any]) -> bool:
        """
        Add ThreatIndicator to the state of this type, if it is for this type.

        Return true if the indicator was used.
        """
        return False

    def load(self, path: pathlib.Path) -> None:
        raise NotImplementedError

    def load_indicators(self, path: pathlib.Path) -> None:
        if not path.exists():
            return
        csv.field_size_limit(path.stat().st_size)  # dodge field size problems
        with path.open("r", newline="") as f:
            for row in csv.reader(f):
                ti = ThreatIndicator.from_row(row)
                self.indicator_state[ti.id] = ti

    def store(self, path: pathlib.Path) -> None:
        raise NotImplementedError

    def store_indicators(self, path: pathlib.Path) -> None:
        with path.open("w+", newline="") as f:
            writer = csv.writer(f)
            for _, i in self.indicator_state.items():
                writer.writerow(i.as_row())


class HashMatcher:
    def match_hash(self, hash: str) -> t.List[SignalMatch]:
        """
        Given a representation of this SignalType, return matches.

        Example - PDQ distance comparison, or MD5 exact comparison
        """
        raise NotImplementedError


class FileMatcher:
    def match_file(self, file: pathlib.Path) -> t.List[SignalMatch]:
        """
        Given a file containing content, return matches.
        """
        raise NotImplementedError


class StrMatcher(FileMatcher):
    def match(self, content: str) -> t.List[SignalMatch]:
        """
        Given a string representation of content, return matches.

        You don't need to implement this if it doesn't make sense for your
        signal type.
        """
        raise NotImplementedError

    def match_file(self, path: pathlib.Path) -> t.List[SignalMatch]:
        return self.match(path.read_text())


class SimpleSignalType(SignalType, HashMatcher):
    """
    Dead simple implementation for loading/storing a SignalType.

    Assumes that the signal type can easily merge on a string.
    """

    INDICATOR_TYPE = ""
    TYPE_TAG = ""

    def __init__(self) -> None:
        self.state: t.Dict[str, SimpleDescriptorRollup] = {}
        self.indicator_state: t.Dict[int, ThreatIndicator] = {}

    def process_descriptor(self, descriptor: ThreatDescriptor) -> bool:
        """
        Add ThreatDescriptor to the state of this type, if it is for this type

        Return true if the true if the descriptor was used.
        """
        if (
            descriptor.indicator_type != self.INDICATOR_TYPE
            or self.TYPE_TAG not in descriptor.tags
        ):
            return False
        old_val = self.state.get(descriptor.raw_indicator)
        if old_val is None:
            self.state[
                descriptor.raw_indicator
            ] = SimpleDescriptorRollup.from_descriptor(descriptor)
        else:
            old_val.merge(descriptor)
        return True

    def process_indicator(self, indicator: ThreatIndicator) -> bool:
        """
        Add ThreatIndicator to the state of this type, if it is for this type

        Return true if the indicator was used.
        """
        if (
            indicator.threat_type != self.INDICATOR_TYPE
            or self.TYPE_TAG not in indicator.tags
        ):
            return False

        self.indicator_state[indicator.id] = indicator

        return True

    def match_hash(self, signal_str: str) -> t.List[SignalMatch]:
        found = self.state.get(signal_str)
        if found:
            return [SignalMatch(found.labels, found.first_descriptor_id)]
        return []

    def load(self, path: pathlib.Path) -> None:
        self.state.clear()
        csv.field_size_limit(path.stat().st_size)  # dodge field size problems
        with path.open("r", newline="") as f:
            for row in csv.reader(f):
                self.state[row[0]] = SimpleDescriptorRollup.from_row(row[1:])

    def store(self, path: pathlib.Path) -> None:
        with path.open("w+", newline="") as f:
            writer = csv.writer(f)
            for k, v in self.state.items():
                writer.writerow((k,) + v.as_row())



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import pdqhash
import pathlib
import numpy as np
from PIL import Image, ImageOps


def pdq_from_file(path: pathlib.Path):
    """
    Given a path to a file return the PDQ Hash string in hex
    Current Supported file types: jpg
    """
    img_pil = Image.open(path)
    image = np.asarray(img_pil)
    hash_vector, quality = pdqhash.compute(image)

    bin_str = "".join([str(x) for x in hash_vector])

    # binary to hex using format string
    # '%0*' is for padding up to ceil(num_bits/4),
    # '%X' create a hex representation from the binary string's integer value
    hex_str = "%0*X" % ((len(bin_str) + 3) // 4, int(bin_str, 2))
    hex_str = hex_str.lower()

    return hex_str, quality



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
A place to put simple helpers that don't seem like they go anywhere else.

If this file starts getting large, break it up.
"""

import re


def class_name_to_human_name(name: str, suffix: str) -> str:
    """Helper to make human-friendly names using a class name as a template"""
    if name.endswith(suffix):
        name = name[: -len(suffix)]
    return camel_case_to_underscore(name)


def camel_case_to_underscore(name: str) -> str:
    """
    Convert name in camel-case notation into lowercase+underscore notation.

    For example, AbcXyz will be converted into abc_xyz.
    TODO: Maybe should live in some kind of common location
    """
    s1 = re.sub(r"(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub(r"([a-z0-9])([A-Z])", r"\1_\2", s1).lower()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
This is an entire copy of a file from ThreatExchange/hashing

TODO: Slim down to only what we need, and switch to using requests
"""

import copy
import datetime
import json

# General Python dependencies
import os
import re
import urllib
import urllib.error
import urllib.parse
import urllib.request


class Net:
    THREAT_DESCRIPTOR = "THREAT_DESCRIPTOR"
    DEFAULT_TE_BASE_URL = "https://graph.facebook.com/v6.0"
    TE_BASE_URL = DEFAULT_TE_BASE_URL
    APP_TOKEN = None

    # This is just a keystroke-saver / error-avoider for passing around
    # post-parameter field names.

    POST_PARAM_NAMES = {
        "indicator": "indicator",  # For submit
        "type": "type",  # For submit
        "descriptor_id": "descriptor_id",  # For update
        "description": "description",
        "share_level": "share_level",
        "status": "status",
        "privacy_type": "privacy_type",
        "privacy_members": "privacy_members",
        "tags": "tags",
        "add_tags": "add_tags",
        "remove_tags": "remove_tags",
        "confidence": "confidence",
        "precision": "precision",
        "review_status": "review_status",
        "severity": "severity",
        "expired_on": "expired_on",
        "first_active": "first_active",
        "last_active": "last_active",
        "related_ids_for_upload": "related_ids_for_upload",
        "related_triples_for_upload_as_json": "related_triples_for_upload_as_json",
        # Legacy : should have been named reactions_to_add, but isn't. :(
        "reactions": "reactions",
        "reactions_to_remove": "reactions_to_remove",
    }

    # ----------------------------------------------------------------
    # Helper method for issuing a GET and returning the JSON payload.
    @classmethod
    def getJSONFromURL(self, url):
        numTries = 0
        while True:
            numTries += 1
            [response, error] = self.tryGET(url)
            if response != None:
                response = response.read()
                # Now make it a string
                response = response.decode("utf-8")
                return json.loads(response)
            elif error.code < 500 or error.code >= 600:
                raise error
            elif numTries > 4:
                raise error

    @classmethod
    def tryGET(self, url):
        try:
            # The timeout is a heuristic
            response = urllib.request.urlopen(url, None, 60)
            return [response, None]
        except urllib.error.HTTPError as e:
            return [None, e]

    # ----------------------------------------------------------------
    # Looks up the "objective tag" ID for a given tag. This is suitable input for the /threat_tags endpoint.

    @classmethod
    def getTagIDFromName(self, tagName, showURLs=False):
        url = (
            self.TE_BASE_URL
            + "/threat_tags"
            + "/?access_token="
            + self.APP_TOKEN
            + "&text="
            + urllib.parse.quote(tagName)
        )
        if showURLs:
            print("URL:")
            print(url)

        response = self.getJSONFromURL(url)

        # The lookup will get everything that has this as a prefix.
        # So we need to filter the results. This loop also handles the
        # case when the results array is empty.
        #
        # Example: when querying for "media_type_video", we want the 2nd one:
        # { "data": [
        #   { "id": "9999338563303771", "text": "media_type_video_long_hash" },
        #   { "id": "9999474908560728", "text": "media_type_video" },
        #   { "id": "9889872714202918", "text": "media_type_video_hash_long" }
        #   ], ...
        # }
        data = response["data"]
        desired = list(filter(lambda o: o["text"] == tagName, data))
        if len(desired) < 1:
            return None
        else:
            return desired[0]["id"]

    # ----------------------------------------------------------------
    # Looks up all metadata for given IDs.
    @classmethod
    def getInfoForIDs(self, ids, **kwargs):
        verbose = kwargs.get("verbose", False)
        showURLs = kwargs.get("showURLs", False)
        includeIndicatorInOutput = kwargs.get("includeIndicatorInOutput", True)

        # Check well-formattedness of descriptor IDs (which may have come from
        # arbitrary data on stdin).
        for id in ids:
            try:
                _ = int(id)
            except ValueError:
                raise Exception('Malformed descriptor ID "%s"' % id)

        # See also
        # https://developers.facebook.com/docs/threat-exchange/reference/apis/threat-descriptor/v6.0
        # for available fields

        url = (
            self.TE_BASE_URL
            + "/?access_token="
            + self.APP_TOKEN
            + "&ids="
            + ",".join(ids)
            + "&fields=raw_indicator,type,added_on,last_updated,confidence,owner,privacy_type,review_status,status,severity,share_level,tags,description,reactions,my_reactions"
        )

        if showURLs:
            print("URL:")
            print(url)

        response = self.getJSONFromURL(url)

        descriptors = []
        for id, descriptor in response.items():
            if includeIndicatorInOutput == False:
                del descriptor["raw_indicator"]
            if verbose:
                print(json.dumps(descriptor))

            tags = descriptor.get("tags", None)
            if tags is None:
                tags = []
            else:
                tags = tags["data"]

            # Canonicalize the tag ordering and simplify the
            # structure to simply an array of tag-texts
            descriptor["tags"] = sorted(tag["text"] for tag in tags)

            if descriptor.get("description") is None:
                descriptor["description"] = ""

            descriptors.append(descriptor)

        return descriptors

    # ----------------------------------------------------------------
    # Gets threat updates for the given privacy group.
    @classmethod
    def getThreatUpdates(self, privacy_group, **kwargs):
        if "next" in kwargs:
            url = kwargs["next"]
        else:
            url = (
                self.TE_BASE_URL
                + "/"
                + str(privacy_group)
                + "/threat_updates/"
                + "?access_token="
                + self.APP_TOKEN
                + "&fields=id,indicator,type,creation_time,last_updated,is_expired,expire_time,tags,status,applications_with_opinions"
            )
            for arg, value in kwargs.items():
                if value is not None:
                    if arg == "additional_tags":
                        url += "&additional_tags=" + ",".join(
                            [str(app) for app in value]
                        )
                    elif arg == "threat_type":
                        url += "&threat_types=" + ",".join(value)
                    else:
                        url += "&" + arg + "=" + str(value)
        return self.getJSONFromURL(url)

    # ----------------------------------------------------------------
    # Returns error message or None.
    # This simply checks to see (client-side) if required fields aren't provided.
    @classmethod
    def validatePostPararmsForSubmit(self, postParams):
        if postParams.get(self.POST_PARAM_NAMES["descriptor_id"]) != None:
            return "descriptor_id must not be specified for submit."

        requiredFields = [
            self.POST_PARAM_NAMES["indicator"],
            self.POST_PARAM_NAMES["type"],
            self.POST_PARAM_NAMES["description"],
            self.POST_PARAM_NAMES["share_level"],
            self.POST_PARAM_NAMES["status"],
            self.POST_PARAM_NAMES["privacy_type"],
        ]

        missingFields = [
            fieldName if postParams.get(fieldName) == None else None
            for fieldName in requiredFields
        ]
        missingFields = [fieldName for fieldName in missingFields if fieldName != None]

        if len(missingFields) == 0:
            return None
        elif len(missingFields) == 1:
            return "Missing field %s" % missingFields[0]
        else:
            return "Missing fields %s" % ",".join(missingFields)

    # ----------------------------------------------------------------
    # Returns error message or None.
    # This simply checks to see (client-side) if required fields aren't provided.
    @classmethod
    def validatePostPararmsForCopy(self, postParams):
        if postParams.get(self.POST_PARAM_NAMES["descriptor_id"]) == None:
            return "Source-descriptor ID must be specified for copy."
        if postParams.get(self.POST_PARAM_NAMES["privacy_type"]) == None:
            return "Privacy type must be specified for copy."
        if postParams.get(self.POST_PARAM_NAMES["privacy_members"]) == None:
            return "Privacy members must be specified for copy."
        return None

    @classmethod
    def reactToThreatDescriptor(
        cls, descriptor_id, reaction, *, showURLs=False, dryRun=False
    ):
        """
        Does a POST to the reactions API.

        See: https://developers.facebook.com/docs/threat-exchange/reference/reacting
        """
        return cls._postThreatDescriptor(
            "/".join(
                (cls.TE_BASE_URL, str(descriptor_id), f"?access_token={cls.APP_TOKEN}")
            ),
            {"reactions": reaction},
            showURLs=showURLs,
            dryRun=dryRun,
        )

    # ----------------------------------------------------------------
    # Does a single POST to the threat_descriptors endpoint.  See also
    # https://developers.facebook.com/docs/threat-exchange/reference/submitting
    @classmethod
    def submitThreatDescriptor(self, postParams, showURLs, dryRun):
        errorMessage = self.validatePostPararmsForSubmit(postParams)
        if errorMessage != None:
            return [errorMessage, None, None]

        url = (
            self.TE_BASE_URL
            + "/threat_descriptors"
            + "/?access_token="
            + self.APP_TOKEN
        )

        return self._postThreatDescriptor(url, postParams, showURLs, dryRun)

    @classmethod
    def copyThreatDescriptor(self, postParams, showURLs, dryRun):
        errorMessage = self.validatePostPararmsForCopy(postParams)
        if errorMessage != None:
            return [errorMessage, None, None]

        # Get source descriptor
        sourceID = postParams["descriptor_id"]
        # Not valid for posting a new descriptor
        del postParams["descriptor_id"]
        sourceDescriptor = self.getInfoForIDs([sourceID], showURLs=showURLs)
        sourceDescriptor = sourceDescriptor[0]

        # Mutate necessary fields
        newDescriptor = copy.deepcopy(sourceDescriptor)
        newDescriptor["indicator"] = sourceDescriptor["raw_indicator"]
        del newDescriptor["raw_indicator"]
        if "tags" in newDescriptor and newDescriptor["tags"] is None:
            del newDescriptor["tags"]

        # The shape is different between the copy-from data (mapping app IDs to
        # reactions) and the post data (just a comma-delimited string of owner-app
        # reactions).
        if "reactions" in newDescriptor:
            del newDescriptor["reactions"]

        # Take the source-descriptor values and overwrite any post-params fields
        # supplied by the caller. Note: Python's dict-update method keeps the old
        # value for a given field name when both old and new are present so we
        # invoke it seemingly 'backward'.
        #
        # Example:
        # * x = {'a': 1, 'b': 2, 'c': 3}
        # * y = {'a': 1, 'b': 9, 'd': 12}
        # * After y.update(x) then x is unchanged and y is
        #       {'a': 1, 'b': 2, 'd': 12, 'c': 3}
        #
        # This means we want newDescriptor.update(postParams)
        newDescriptor.update(postParams)

        # Get rid of fields like last_upated from the source descriptor which
        # aren't valid for post
        postParams = {}
        for key, value in newDescriptor.items():
            if self.POST_PARAM_NAMES.get(key) != None:
                postParams[key] = value

        return self.submitThreatDescriptor(postParams, showURLs, dryRun)

    # ----------------------------------------------------------------
    # Code-reuse for submit and update
    @classmethod
    def _postThreatDescriptor(self, url, postParams, showURLs, dryRun):
        for key, value in postParams.items():
            url += "&%s=%s" % (key, urllib.parse.quote(str(value)))
        if showURLs:
            print()
            print("URL:")
            print(url)
        if dryRun:
            print("Not doing POST since --dry-run.")
            return [None, None, ""]

        # Encode the inputs to the POST
        header = {"Content-Type": "text/json", "charset": "utf-8"}
        # This is a string
        data = urllib.parse.urlencode(postParams)
        # Turn it into a Python bytes object
        data = data.encode("ascii")

        # Do the POST
        try:
            response = urllib.request.urlopen(url, data)

            # Decode the outputs from the POST
            # This is a Python 'bytes'
            response = response.read()
            # Now make it a string
            response = response.decode("utf-8")
            responseBody = json.loads(response)
            responseCode = None

            return [None, None, responseBody]

        except urllib.error.HTTPError as e:
            responseBody = json.loads(e.read().decode("utf-8"))
            return [None, e, responseBody]
